﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.DatastreamAnalysis {
  internal enum DatastreamAnalysisOptimizerAction { None, Prepare, Start, Stop, Pause }

  [StorableClass]
  [Item("DatastreamAnalysis Optimizer", "The main loop for evaluating ensemble models against a incoming datastream of time series fashion.")]
  [Creatable("Algorithms")]
  public class DatastreamAnalysisOptimizer : NamedItem, IOptimizer, IStorableContent {

    // TODO: introduce members -> ensembles, datastream


    // VAR 1: datastream = problem data
    [Storable]
    private IRegressionProblemData problemData;

    public IRegressionProblemData ProblemData {
      get { return problemData; }
      set {
        if (value == null || value == problemData)
          return;

        problemData = value;
      }
    }

    // VAR 2: datastream = external source (webservice, AMQP-Queue, etc.)
    // TODO

    public DatastreamAnalysisOptimizer() {
      Name = "Datastream Analysis";
    }

    [StorableConstructor]
    protected DatastreamAnalysisOptimizer(bool deserializing) : base(deserializing) { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // nothing to do in here
    }

    protected DatastreamAnalysisOptimizer(DatastreamAnalysisOptimizer original, Cloner cloner) : base(original, cloner) {
      // TODO: clone members (ensembles, datastream)
      this.ProblemData = (IRegressionProblemData)original.ProblemData.Clone(cloner);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return  new DatastreamAnalysisOptimizer(this, cloner);
    }


    // TODO: introduce EventHandler for EnsemblesChanged, DatastreamChanged

    public EventHandler ProblemDataChanged;
    private void OnProblemDataChanged() {
      var changed = ProblemDataChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }


    public void Prepare() {
      Prepare(false);
    }

    public void Start() {
      if ((ExecutionState != ExecutionState.Prepared) && (ExecutionState != ExecutionState.Paused))
        throw new InvalidOperationException(string.Format("Start not allowed in execution state \"{0}\".", ExecutionState));
    }

    public void Pause() {
      if (ExecutionState != ExecutionState.Started)
        throw new InvalidOperationException(string.Format("Pause not allowed in execution state \"{0}\".", ExecutionState));
    }

    public void Stop() {
      if ((ExecutionState != ExecutionState.Started) && (ExecutionState != ExecutionState.Paused))
        throw new InvalidOperationException(string.Format("Stop not allowed in execution state \"{0}\".", ExecutionState));
    }

    public void Prepare(bool clearRuns) {
      if ((ExecutionState != ExecutionState.Prepared) && (ExecutionState != ExecutionState.Paused) && (ExecutionState != ExecutionState.Stopped))
        throw new InvalidOperationException(string.Format("Prepare not allowed in execution state \"{0}\".", ExecutionState));
    }


    [Storable]
    private TimeSpan executionTime;
    public TimeSpan ExecutionTime {
      get { return executionTime; }
      private set {
        executionTime = value;
        OnExecutionTimeChanged();
      }
    }

    public RunCollection Runs {
      get {
        return new RunCollection();
      }
    }

    public IEnumerable<IOptimizer> NestedOptimizers { get; }
    public string Filename { get; set; }

    public new static Image StaticItemImage
    {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Event; }
    }

    public override Image ItemImage {
      get {
        if (ExecutionState == ExecutionState.Prepared)
          return HeuristicLab.Common.Resources.VSImageLibrary.BatchRunPrepared;
        else if (ExecutionState == ExecutionState.Started)
          return HeuristicLab.Common.Resources.VSImageLibrary.BatchRunStarted;
        else if (ExecutionState == ExecutionState.Paused) return HeuristicLab.Common.Resources.VSImageLibrary.BatchRunPaused;
        else if (ExecutionState == ExecutionState.Stopped)
          return HeuristicLab.Common.Resources.VSImageLibrary.BatchRunStopped;
        else return base.ItemImage;
      }
    }

    #region Events

    public ExecutionState ExecutionState { get; private set; }
    public event EventHandler ExecutionStateChanged;
    public event EventHandler ExecutionTimeChanged;
    public event EventHandler Prepared;
    public event EventHandler Started;
    public event EventHandler Paused;
    public event EventHandler Stopped;
    public event EventHandler<EventArgs<Exception>> ExceptionOccurred;


    #region event firing

    protected override void OnNameChanged() {
      base.OnNameChanged();
    }

    private void OnExecutionStateChanged() {
      EventHandler handler = ExecutionStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnExecutionTimeChanged() {
      EventHandler handler = ExecutionTimeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnPrepared() {
      ExecutionState = ExecutionState.Prepared;
      EventHandler handler = Prepared;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnStarted() {
      ExecutionState = ExecutionState.Started;
      EventHandler handler = Started;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnPaused() {
      ExecutionState = ExecutionState.Paused;
      EventHandler handler = Paused;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnStopped() {
      ExecutionState = ExecutionState.Stopped;
      EventHandler handler = Stopped;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void OnExceptionOccurred(Exception exception) {
      EventHandler<EventArgs<Exception>> handler = ExceptionOccurred;
      if (handler != null) handler(this, new EventArgs<Exception>(exception));
    }

    #endregion


    #endregion

  }
}
